/*
 *  config_file.c - manages config file access.
 *	part of galculator
 *  	(c) 2002-2003 Simon Floery (simon.floery@gmx.at)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
/* to add a new preference, add it to 
	- s_preferences in config_file.h
	- to config_file_get_default_prefs
	- to struct prefs_list
*/

#include <stdio.h>
#include <string.h>

#include "config_file.h"
#include "general_functions.h"

s_preferences prefs;
s_prefs_entry prefs_list[26] = {
	{"display_bkg_color", &(prefs.bkg_color), STRING, "prefs_bkg_color", get_button_color, set_button_color},
	{"display_result_font", &(prefs.result_font), STRING, "prefs_result_font", get_button_label, set_button_label},
	{"display_result_color", &(prefs.result_color), STRING, "prefs_result_color", get_button_color, set_button_color},
	{"display_module_font", &(prefs.mod_font), STRING, "prefs_mod_font", get_button_label, set_button_label},
	{"display_module_active_color", &(prefs.act_mod_color), STRING, "prefs_act_mod_color", get_button_color, set_button_color},
	{"display_module_inactive_color", &(prefs.inact_mod_color), STRING, "prefs_inact_mod_color", get_button_color, set_button_color},
	{"display_module_number", &(prefs.vis_number), BOOLEAN, "prefs_vis_number", get_checkbutton, set_checkbutton},
	{"display_module_angle", &(prefs.vis_angle), BOOLEAN, "prefs_vis_angle", get_checkbutton, set_checkbutton},
	{"display_module_notation", &(prefs.vis_notation), BOOLEAN, "prefs_vis_notation", get_checkbutton, set_checkbutton},
	{"display_module_arith", &(prefs.vis_arith), BOOLEAN, "prefs_vis_arith", get_checkbutton, set_checkbutton},
	{"display_module_open", &(prefs.vis_bracket), BOOLEAN, "prefs_vis_bracket", get_checkbutton, set_checkbutton},
	{"custom_button_font", &(prefs.custom_button_font), BOOLEAN, "prefs_custom_button_font", get_checkbutton, set_checkbutton},
	{"button_font", &(prefs.button_font), STRING, "prefs_button_font", get_button_label, set_button_label},
	{"button_width", &(prefs.button_width), INTEGER, "prefs_button_width", get_spinbutton, set_spinbutton},
	{"button_height", &(prefs.button_height), INTEGER, "prefs_button_height", get_spinbutton, set_spinbutton},
	{"function_button_group", &(prefs.vis_funcs), BOOLEAN, "prefs_vis_funcs", get_checkbutton, set_checkbutton},
	{"logic_button_group", &(prefs.vis_logic), BOOLEAN, "prefs_vis_logic", get_checkbutton, set_checkbutton},
	{"dispctrl_button_group", &(prefs.vis_dispctrl), BOOLEAN, "prefs_vis_dispctrl", get_checkbutton, set_checkbutton},
	{"default_number_base", &(prefs.def_number), INTEGER, "prefs_def_number", get_optmenu, set_optmenu},
	{"default_angle_base", &(prefs.def_angle), INTEGER, "prefs_def_angle", get_optmenu, set_optmenu},
	{"default_notation_mode", &(prefs.def_notation), INTEGER, "prefs_def_notation", get_optmenu, set_optmenu},
	{"remembers_display", &(prefs.rem_display), BOOLEAN, "prefs_rem_display", get_checkbutton, set_checkbutton},
	{"remembered_value", &(prefs.rem_value), STRING, NULL, NULL, NULL},
	{"show_menu_bar", &(prefs.show_menu), BOOLEAN, "prefs_show_menu", get_checkbutton, set_checkbutton},
	{"show_status_bar", &(prefs.show_status), BOOLEAN, "prefs_show_status", get_checkbutton, set_checkbutton},
	{NULL, NULL, 0, NULL, NULL, NULL}
};


/*
 * config_file_get_default_prefs - initialize ALL members of the given s_preferences
 *	structure with default values.
 */
static void config_file_get_default_prefs (s_preferences *this_prefs)
{
	// 1st pref page
	this_prefs->bkg_color = g_strdup (DEFAULT_BKG_COLOR);
	this_prefs->result_font = g_strdup (DEFAULT_RESULT_FONT);
	this_prefs->result_color = g_strdup (DEFAULT_RESULT_COLOR);
	this_prefs->mod_font = g_strdup (DEFAULT_MOD_FONT);
	this_prefs->act_mod_color = g_strdup (DEFAULT_ACT_MOD_COLOR);
	this_prefs->inact_mod_color = g_strdup (DEFAULT_INACT_MOD_COLOR);
	this_prefs->vis_number = DEFAULT_VIS_NUMBER;
	this_prefs->vis_angle = DEFAULT_VIS_ANGLE;
	this_prefs->vis_notation = DEFAULT_VIS_NOTATION;
	this_prefs->vis_arith = DEFAULT_VIS_ARITH;
	this_prefs->vis_bracket = DEFAULT_VIS_BRACKET;
	
	// 2nd pref page
	this_prefs->custom_button_font = DEFAULT_CUSTOM_BUTTON_FONT;
	this_prefs->button_font = g_strdup (DEFAULT_BUTTON_FONT);
	this_prefs->button_width = DEFAULT_BUTTON_WIDTH;
	this_prefs->button_height = DEFAULT_BUTTON_HEIGHT;
	this_prefs->vis_funcs = DEFAULT_VIS_FUNCS;
	this_prefs->vis_logic = DEFAULT_VIS_LOGIC;
	this_prefs->vis_dispctrl = DEFAULT_VIS_DISPCTRL;
	
	// 3rd pref page
	this_prefs->def_number = DEFAULT_NUMBER;
	this_prefs->def_angle = DEFAULT_ANGLE;
	this_prefs->def_notation = DEFAULT_NOTATION;
	this_prefs->rem_display = DEFAULT_REM_DISPLAY;
	this_prefs->rem_value = g_strdup (DEFAULT_REM_VALUE);
	this_prefs->show_menu = DEFAULT_SHOW_MENU;
	this_prefs->show_status = DEFAULT_SHOW_STATUS;
}

/*
 * config_file_set_prefs - find the key in the s_prefs_entry struct to retrieve
 *	the variable to set. The way to set this variable is given by key_type.
 * 	used by config_file_read.
 */

void config_file_set_prefs (s_preferences *this_prefs, char *key, char *value)
{
	int						*int_var, counter=0;
	char 					**string_var;
	gboolean				*bool_var;
	void					*this_var;
	
	while (prefs_list[counter].key != NULL)
	{
		if (g_ascii_strcasecmp (key, prefs_list[counter].key) == 0) break;	
		counter++;
	}
	
	if (prefs_list[counter].key == NULL)
	{
		fprintf (stderr, "[%s] config file: ignoring unknown entry %s=%s. %s\n", PROG_NAME, key, value, BUG_REPORT);
		return;
	}
	
	this_var = prefs_list[counter].variable;
	switch (prefs_list[counter].key_type)
	{
		case STRING:
			string_var = this_var;
			g_free (*string_var);
			*string_var = g_strdup (value);
			break;
		case BOOLEAN:
			bool_var = this_var;
			if (g_ascii_strcasecmp (value, "true")  == 0) *bool_var = TRUE;
			else if (g_ascii_strcasecmp (value, "false")  == 0) *bool_var = FALSE;
			else fprintf (stderr, "[%s] config file: %s has to be TRUE or FALSE. Using defaults. %s\n", PROG_NAME, key, BUG_REPORT);
			break;
		case INTEGER:
			int_var = this_var;
			*int_var = (int) g_ascii_strtod (value, NULL);
			break;
		default:
			fprintf (stderr, "[%s] config file: ignoring unknown key_type in config structure. %s\n", PROG_NAME, BUG_REPORT);
	}
}

/*
 * config_file_read - open/read/close. values are saved to global variable prefs!
 */

void config_file_read (char *filename)
{
	char			line[MAX_FILE_LINE_LENGTH], *key, *value;
	FILE			*this_file;
	
	this_file = fopen (filename, "r");
	config_file_get_default_prefs (&prefs);
	if (this_file != NULL)
	{
		while (fgets (line, MAX_FILE_LINE_LENGTH, this_file) != NULL)
		{
			if (line[0] != '#') {
				key = line;
				value = index (line, '=');
				*value = '\0';
				value++;
				key = g_strstrip(key);
				value = g_strstrip(value);
				config_file_set_prefs (&prefs, key, value);
			}
		}
		fclose (this_file);
	}
	else fprintf (stderr, "[%s] config file: couldn't open config file %s for reading. \
Nothing to worry about if you are using %s for the first time. Using defaults.\n", PROG_NAME, filename, PROG_NAME);
}

/*
 * config_file_write - open/write/close.
 */

void config_file_write (char *filename, s_preferences this_prefs)
{
	int						*int_var, counter=0;
	char 					**string_var;
	gboolean				*bool_var;
	void					*this_var;
	char 					*line=NULL;
	FILE					*this_file;
	
	this_file = fopen (filename, "w+");
	
	if (this_file != NULL)
	{
		while (prefs_list[counter].key != NULL)
		{
			this_var = prefs_list[counter].variable;
			switch (prefs_list[counter].key_type)
			{
				case STRING:
					string_var = this_var;
					line = g_strdup_printf ("%s=%s\n", prefs_list[counter].key, *string_var);
					break;
				case BOOLEAN:
					bool_var = this_var;
					if (*bool_var == TRUE) line = g_strdup_printf ("%s=true\n", prefs_list[counter].key);
					else if (*bool_var == FALSE) line = g_strdup_printf ("%s=false\n", prefs_list[counter].key);
					else fprintf (stderr, "[%s] config file: strange boolean when writing. skipping. %s\n", PROG_NAME, BUG_REPORT);
					break;
				case INTEGER:
					int_var = this_var;
					line = g_strdup_printf ("%s=%i\n", prefs_list[counter].key, *int_var);
					break;
				default:
					line = g_strdup_printf ("#%s=???\n", prefs_list[counter].key);
					fprintf (stderr, "[%s] config file: ignoring unknown key_type in config structure. %s\n", PROG_NAME, BUG_REPORT);
			}
			fputs (line, this_file);
			g_free (line);
			counter ++;
		}
		fclose (this_file);
	}	
	else fprintf (stderr, "[%s] config file: couldn't save/write to configuration file %s.\n", PROG_NAME, filename);
}
