/*
 *  general_functions.c - this and that.
 *	part of galculator
 *  	(c) 2002-2003 Simon Floery (simon.floery@gmx.at)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gtk/gtk.h>

#include "galculator.h"
#include "general_functions.h"
#include "math_functions.h"
#include "calc_basic.h"
#include "display.h"
#include "config_file.h"

#include "interface.h"
#include "support.h"

GtkStatusbar 		*statusbar;
double				memory_value=0;

// assume TRUE for all other bases/modes!

s_active_buttons active_buttons[] = {\
	{"button_2", {TRUE, TRUE, TRUE, FALSE}, {TRUE, TRUE}}, \
	{"button_3", {TRUE, TRUE, TRUE, FALSE}, {TRUE, TRUE}}, \
	{"button_4", {TRUE, TRUE, TRUE, FALSE}, {TRUE, TRUE}}, \
	{"button_5", {TRUE, TRUE, TRUE, FALSE}, {TRUE, TRUE}}, \
	{"button_6", {TRUE, TRUE, TRUE, FALSE}, {TRUE, TRUE}}, \
	{"button_7", {TRUE, TRUE, TRUE, FALSE}, {TRUE, TRUE}}, \
	{"button_8", {TRUE, TRUE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_9", {TRUE, TRUE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_a", {FALSE, TRUE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_b", {FALSE, TRUE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_c", {FALSE, TRUE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_d", {FALSE, TRUE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_e", {FALSE, TRUE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_f", {FALSE, TRUE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_econst", {TRUE, FALSE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_piconst", {TRUE, FALSE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_ee", {TRUE, FALSE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_sin", {TRUE, FALSE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_cos", {TRUE, FALSE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_tan", {TRUE, FALSE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_reci", {TRUE, FALSE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_point", {TRUE, FALSE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_sign", {TRUE, FALSE, FALSE, FALSE}, {TRUE, TRUE}}, \
	{"button_paropen", {TRUE, TRUE, TRUE, TRUE}, {TRUE, FALSE}}, \
	{"button_parclose", {TRUE, TRUE, TRUE, TRUE}, {TRUE, FALSE}}, \
	{NULL}\
};

void statusbar_init (GtkWidget *a_parent_widget)
{
	int		context_id;
	
	statusbar = (GtkStatusbar*) lookup_widget (a_parent_widget, "statusbar");
	context_id = gtk_statusbar_get_context_id (statusbar, "standard statusbar");
	gtk_statusbar_push (statusbar, context_id, g_strdup_printf ("welcome to %s v%s", PROG_NAME, PROG_VERSION));
}

double error_unsupported_inv (double dummy)
{
	error_message ("unsupported inverse");
	return dummy;
}

double error_unsupported_hyp (double dummy)
{
	error_message ("unsupported hyperbolic function");
	return dummy;
}

void error_message (char *message)
{
	int		context_id;
	
	fprintf (stderr, "[%s] %s. %s\n", PROG_NAME, message, BUG_REPORT);
	context_id = gtk_statusbar_get_context_id (statusbar, "error message");
	gtk_statusbar_push (statusbar, context_id, message);
}

/* the last entered number is removed. it is impossible to remove the last operation:
 * this would be a quite difficult taks, as every possible computation is done asap.
 * so if you entered 1+2- and you want to correct to 1+2/ calc_basic would have already
 * calculated 1+2=3 and set the tree to 3-. you would have to make a backup of the tree!*/

void clear ()
{
	display_result_set (DEFAULT_REM_VALUE);
}

/* clear all: display ("0"), calc_tree ... */

void all_clear ()
{
	extern gboolean 			rpn_have_result;
	extern s_current_status 	current_status;

	clear();
	if (current_status.notation == CS_PAN) calc_tree_free();
	else 
	{
		calc_rpn_free();
		rpn_have_result=FALSE;
	}
	display_module_bracket_label_update (RESET);
}

void memory_update_tooltips ()
{
	static GtkWidget	*button_ms=NULL, *button_mr=NULL, *button_mplus=NULL;
	static char 		*button_ms_format=NULL, *button_mr_format=NULL, \
						*button_mplus_format=NULL;
	GtkTooltipsData		*tooltips_data;
	extern GtkWidget	*main_window;
	
	if (button_ms == NULL) {
		button_ms = lookup_widget (main_window, "button_MS");
		button_mr = lookup_widget (main_window, "button_MR");
		button_mplus = lookup_widget (main_window, "button_Mplus");
	}
	if (button_ms_format == NULL) {
		tooltips_data = gtk_tooltips_data_get (button_ms);
		button_ms_format = g_strdup (tooltips_data->tip_text);
		tooltips_data = gtk_tooltips_data_get (button_mr);
		button_mr_format = g_strdup (tooltips_data->tip_text);
		tooltips_data = gtk_tooltips_data_get (button_mplus);
		button_mplus_format = g_strdup (tooltips_data->tip_text);
	}
	tooltips_data = gtk_tooltips_data_get (button_ms);
	g_free (tooltips_data->tip_text);
	tooltips_data->tip_text = g_strdup_printf (button_ms_format, memory_value);
	
	tooltips_data = gtk_tooltips_data_get (button_mr);
	g_free (tooltips_data->tip_text);
	tooltips_data->tip_text = g_strdup_printf (button_mr_format, memory_value);
		
	tooltips_data = gtk_tooltips_data_get (button_mplus);
	g_free (tooltips_data->tip_text);
	tooltips_data->tip_text = g_strdup_printf (button_mplus_format, memory_value);
}

void memory_save ()
{
	memory_value = display_result_get_as_double ();
	memory_update_tooltips(memory_value);
}

void memory_read ()
{
	display_result_set_double (memory_value);
}

void memory_add2mem ()
{
	memory_value += display_result_get_as_double ();
	memory_update_tooltips(memory_value);
	/* display this value ? */
}

/*
 * helper function for set_button_group_size
 */

void set_table_child_size (gpointer data, gpointer user_data)
{
	s_point			*size;
	GtkTableChild	*table_child;
	
	size = user_data;				/* dereference */
	table_child = data;
	gtk_widget_set_size_request (table_child->widget, size->x, size->y);
}

/*
 * run through all children of the given table and resize them (done by 
 * set_table_child_size)
 */

void set_button_group_size (GtkWidget *a_parent_widget, char *table_name, int width, int height)
{
	GtkTable	*this_table;
	s_point		size;
	
	size.x = width;
	size.y = height;
	this_table = (GtkTable *) lookup_widget (a_parent_widget, table_name);
	g_list_foreach (this_table->children, set_table_child_size, &size);
	//gtk_window_resize ((GtkWindow *)gtk_widget_get_toplevel(a_parent_widget), 1, 1);
}

/*
 * the same as for set_button_group_size. helper function for set_button_group_font
 */

void set_table_child_font (gpointer data, gpointer user_data)
{
	PangoFontDescription	*font;
	GtkTableChild			*table_child;
	
	font = user_data;				/* dereference */
	table_child = data;
	/* ATTENTION: we don't want to change the button's font but the font of the button's label! */
	gtk_widget_modify_font (gtk_bin_get_child ((GtkBin *)(table_child->widget)), font);
}

/*
 * new font for the labels of the buttons of the given table
 */ 

void set_button_group_font (GtkWidget *a_parent_widget, char *table_name, char *font_string)
{
	GtkTable				*this_table;
	PangoFontDescription	*pango_font;

	pango_font = pango_font_description_from_string (font_string);
	this_table = (GtkTable *) lookup_widget (a_parent_widget, table_name);
	g_list_foreach (this_table->children, set_table_child_font, pango_font);
	pango_font_description_free(pango_font);
}

/* axtof: convert string to float 
 *	works up to base 19. only for integers!
 *  a number is called negative, if its msb is set!
 */

double axtof (char *bin_string, int base, int display_length)
{
	double 		return_value=0;
	int			counter, sign_length=0;
	
	/* according to man strtod, inf should be there in every case */
	if (strstr (g_ascii_strdown (bin_string, -1), "inf") != NULL) return INFINITY;
	for (counter = strlen (bin_string) - 1; counter >= 0; counter--) 
	{
		if (bin_string[counter] - '0' < 10) \
			return_value += (bin_string[counter] - '0') * pow (base, strlen (bin_string) - 1 - counter);
		else if (bin_string[counter] - 'A' < 10) \
			return_value += (bin_string[counter] - 'A' + 10) * pow (base, strlen (bin_string) - 1 - counter);
		else fprintf (stderr, "[%s] failed to convert char %c in axtof. %s\n", PROG_NAME, bin_string[counter], BUG_REPORT);
	}
	
	/* handle negative numbers. */
	/* determine, how much bits a sign of this number base stands for */
	while (pow (2, sign_length) < base) sign_length++;
	/* if most significant bit is set, its a negative number. using
	 * pow (base, display_length - 1) as limit is WRONG.
	 */
	if (return_value >= pow (2, sign_length*display_length - 1)) \
		return_value = (-1) * (pow (base, display_length) - return_value);
	return return_value;
}

/* rem: my own remainder function to deal with precision problems */

int rem (double x, long long int y)
{
	return x - ((long long int)(x/((double)y)))*y;
}

/* ftoax - number to string conversion.
 *	works up to base 19. only for integers!
 *  see axtof for details about what is a negative number.
 */

char *ftoax (double x, int base, int display_length)
{
	char			*return_string;
	int				length=0, counter, remainder, sign_length=0;
	double			localx;
	
	/* determine, how much bits a sign for this number base stands for */
	while (pow (2, sign_length) < base) sign_length++;
	/* handle huge values --> infinity */
	if (x < (-1)*pow (2, display_length*sign_length - 1)) return g_strdup(MY_INFINITY_STRING);
	if (x >= pow (2, display_length*sign_length - 1)) return g_strdup(MY_INFINITY_STRING);
		
	/* handle negative numbers */
	if (x < 0) x = pow (2, display_length*sign_length) + x;
	
	/* doing it this way and not with logs as this is much more numerical stable */
	localx = x;
	while ((localx=floor(localx/(double)base)) >= 1) length++;
	length+=2;
	
	localx = x;
	return_string = (char *) malloc (length * sizeof(char));
	return_string [length-1] = '\0';
	for (counter = length-2; counter >= 0; counter--)
	{
		remainder = rem (localx, base);
		if (remainder < 10) return_string[counter] = '0' + remainder;
		else if (remainder < 20) return_string[counter] = 'A' + remainder - 10;
		else fprintf (stderr, "[%s] failed to convert %f in ftoax. %s\n", PROG_NAME, x, BUG_REPORT);
		localx = floor (localx/(double)base);
	}
	return return_string;
}

/*
 * preference dialog sets. these are handler for the big configuration struct.
 */

void set_button_label (GtkWidget *a_parent_widget, char *button_name, void *new_label)
{
	GtkButton	*button;
	char 		**string_var;
	
	string_var = new_label;	
	button = (GtkButton *) lookup_widget (a_parent_widget, button_name);
	gtk_button_set_label (button, *string_var);	
}

void set_checkbutton (GtkWidget *a_parent_widget, char *checkbutton_name, void *is_active)
{
	GtkToggleButton		*toggle_button;
	gboolean			*bool_var;
	
	bool_var = is_active;
	toggle_button = (GtkToggleButton *) lookup_widget (a_parent_widget, checkbutton_name);	
	gtk_toggle_button_set_active (toggle_button, *bool_var);
}

void set_spinbutton (GtkWidget *a_parent_widget, char *spinbutton_name, void *value)
{
	GtkSpinButton	*spin_button;
	int				*int_var;
	
	int_var = value;
	spin_button = (GtkSpinButton *) lookup_widget (a_parent_widget, spinbutton_name);
	gtk_spin_button_set_value (spin_button, (double) *int_var);
}

void set_optmenu (GtkWidget *a_parent_widget, char *optmenu_name, void *index)
{
	GtkOptionMenu	*opt_menu;
	int				*int_var;
	
	int_var = index;	
	opt_menu = (GtkOptionMenu *) lookup_widget (a_parent_widget, optmenu_name);
	gtk_option_menu_set_history (opt_menu, *int_var);
}

/*
 * The callback fills the given widget with a rectangle in its foreground color.
 */

gboolean da_expose_event_cb(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
	GtkRequisition	requ;
	
	gtk_widget_size_request (widget->parent->parent, &requ);
	gdk_draw_rectangle (widget->window, widget->style->fg_gc[0], TRUE, 0, 0,\
			requ.width, requ.height);
	return TRUE;
}

/*
 * assembling my own colored button. as gnomecolorpicker is in libgnomeui, i do
 * it myself. remove the current label widget in button and put there a 
 * gtkdrawingarea. the drawing is done in da_expose_event_cb. The color is
 * set via widget->style->"fgs"
 */

void set_button_color (GtkWidget *a_parent_widget, char *button_name, void *color_string)
{
	GtkWidget		*button, *label, *da, *viewport;
	GtkRequisition	requ;
	char 			**string_var;
	GdkColor		color;

	// dereference
	string_var = color_string;
	gdk_color_parse (*string_var, &color);
	
	button = lookup_widget (a_parent_widget, button_name);
	gtk_widget_size_request (button, &requ);
	
	label = gtk_bin_get_child((GtkBin *)button);
	gtk_container_remove ((GtkContainer *) button, label);
	
	viewport = gtk_viewport_new (NULL, NULL);
	gtk_container_add ((GtkContainer *) button, viewport);
	//gtk_container_set_border_width ((GtkContainer *) viewport, 1);
	gtk_widget_show (viewport);
	
	da = gtk_drawing_area_new ();
	gtk_container_add ((GtkContainer *) viewport, da);
	gtk_widget_show (da);
	
	gtk_widget_modify_fg (da, 0, &color);
	
	gtk_widget_set_size_request (button, requ.width, requ.height);
	g_signal_connect (G_OBJECT (da), "expose_event", \
		G_CALLBACK (da_expose_event_cb), NULL);
}

/*
 * convert given GdkColor to a string so that gdk_color_parse gives the 
 * same color again.
 */

char *gdk_color_to_string (GdkColor color)
{
	return g_strdup_printf ("#%04X%04X%04X", color.red, color.green, color.blue);
}

GtkWidget *show_font_dialog (char *title, GtkButton *button)
{
	GtkWidget		*font_dialog;
	
	font_dialog = create_font_dialog ();
	gtk_window_set_title ((GtkWindow *) font_dialog, title);
	gtk_font_selection_dialog_set_font_name ((GtkFontSelectionDialog *) font_dialog, \
		gtk_button_get_label (button));
	gtk_widget_hide (((GtkFontSelectionDialog *)font_dialog)->apply_button);
	gtk_widget_show (font_dialog);
	return font_dialog;
}

GtkWidget *show_color_dialog (char *title, GtkButton *button)
{
	GtkWidget		*color_dialog, *da;
	GdkColor		color;
	GtkRcStyle		*style;
	
	da = gtk_bin_get_child ((GtkBin *) gtk_bin_get_child ((GtkBin *) button));
	style = gtk_widget_get_modifier_style (da);
	color = style->fg[GTK_STATE_NORMAL];
	
	color_dialog = create_color_dialog ();
	gtk_window_set_title ((GtkWindow *) color_dialog, title);
	gtk_color_selection_set_current_color ((GtkColorSelection *)((GtkColorSelectionDialog *)color_dialog)->colorsel, \
		&color);
	gtk_widget_hide (((GtkColorSelectionDialog *)color_dialog)->help_button);
	gtk_widget_show (color_dialog);
	return color_dialog;
}

/*
 * get preferences. used in update_all via bg configuration struct.
 */

void get_button_color (GtkWidget *a_parent_widget, char *button_name, void **string_color)
{
	GtkBin		*button;
	GtkWidget	*da;
	GdkColor	color;
	GtkRcStyle	*style;
	
	button = (GtkBin *) lookup_widget (a_parent_widget, button_name);
	da = gtk_bin_get_child((GtkBin *) gtk_bin_get_child (button));
	style = gtk_widget_get_modifier_style (da);
	color = style->fg[GTK_STATE_NORMAL];
	*string_color = (void *) gdk_color_to_string (color);
}

void get_button_label (GtkWidget *a_parent_widget, char *button_name, void **label)
{
	GtkButton	*button;
	
	button = (GtkButton *) lookup_widget (a_parent_widget, button_name);
	*label = (void *) g_strdup (gtk_button_get_label (button));
}

void get_checkbutton (GtkWidget *a_parent_widget, char *name, void **is_active)
{
	GtkToggleButton		*button;
	
	button = (GtkToggleButton *) lookup_widget (a_parent_widget, name);
	*is_active = (void *) gtk_toggle_button_get_active (button);
}

void get_spinbutton (GtkWidget *a_parent_widget, char *name, void **value)
{
	GtkSpinButton		*button;
	
	button = (GtkSpinButton *) lookup_widget (a_parent_widget, name);
	*value = (void *) gtk_spin_button_get_value_as_int (button);
}

void get_optmenu (GtkWidget *a_parent_widget, char *name, void **index)
{
	GtkOptionMenu		*optmenu;
	
	optmenu = (GtkOptionMenu *) lookup_widget (a_parent_widget, name);
	*index = (void *) gtk_option_menu_get_history (optmenu);
}

/*
 * "apply" preference dialog.
 */

void update_all (GtkWidget *a_parent_widget, s_preferences prefs)
{
	GtkCheckMenuItem		*show_menubar_item;
	char					*button_font;
	
	display_update_result_tag (prefs.result_font, prefs.result_color);
	display_set_bkg_color (prefs.bkg_color);
	
	// don't forget the buttons right from the display
	set_button_group_size (a_parent_widget, "table_func", prefs.button_width, prefs.button_height);
	set_button_group_size (a_parent_widget, "table_bin", prefs.button_width, prefs.button_height);
	set_button_group_size (a_parent_widget, "table_standard", prefs.button_width, prefs.button_height);
	set_button_group_size (a_parent_widget, "table_dispctrl", prefs.button_width, prefs.button_height);
	
	if (prefs.custom_button_font == TRUE) button_font = g_strdup (prefs.button_font);
	else button_font = g_strdup ("");
	set_button_group_font (a_parent_widget, "table_func", button_font);
	set_button_group_font (a_parent_widget, "table_bin", button_font);	
	set_button_group_font (a_parent_widget, "table_standard", button_font);	
	set_button_group_font (a_parent_widget, "table_dispctrl", button_font);
	
	set_widget_visibility (a_parent_widget, "table_bin", prefs.vis_logic);
	set_widget_visibility (a_parent_widget, "table_func", prefs.vis_funcs);
	set_widget_visibility (a_parent_widget, "table_dispctrl", prefs.vis_dispctrl);
	
	set_widget_visibility (a_parent_widget, "menubar", prefs.show_menu);
	show_menubar_item = (GtkCheckMenuItem *) lookup_widget (a_parent_widget, "show_menubar1");
	gtk_check_menu_item_set_active (show_menubar_item, prefs.show_menu);
	set_widget_visibility (a_parent_widget, "statusbar", prefs.show_status);
	
	gtk_window_resize ((GtkWindow *)gtk_widget_get_toplevel(a_parent_widget), 1, 1);
}

/*
 * (un) hide a widget
 */ 
void set_widget_visibility (GtkWidget *a_parent_widget, char *widget_name, gboolean visible)
{
	GtkWidget		*widget;
	
	widget = lookup_widget (a_parent_widget, widget_name);
	if (visible) gtk_widget_show_all (widget);
	else gtk_widget_hide_all (widget);
}

/*
 * this function changes the foreground color for all states to the given string
 */

void gtk_widget_really_modify_fg (GtkWidget *widget, GdkColor color)
{
    gtk_widget_modify_fg (widget, GTK_STATE_NORMAL, &color);
	gtk_widget_modify_fg (widget, GTK_STATE_ACTIVE, &color);
	gtk_widget_modify_fg (widget, GTK_STATE_PRELIGHT, &color);
	gtk_widget_modify_fg (widget, GTK_STATE_SELECTED, &color);
	gtk_widget_modify_fg (widget, GTK_STATE_INSENSITIVE, &color);
}

void update_active_buttons (GtkWidget *parent_widget, int number_base, int notation_mode)
{
	int				counter=0;
	GtkWidget		*current_button;
	gboolean		state;
	
	while (active_buttons[counter].button_name != NULL) {
		current_button = lookup_widget (parent_widget, active_buttons[counter].button_name);
		state = active_buttons[counter].number_active[number_base] & \
			active_buttons[counter].notation_active[notation_mode];
		gtk_widget_set_sensitive (current_button, state);
		counter++;
	}
}
