/*
 *  main.c
 *	part of galculator
 *  	(c) 2002-2003 Simon Floery (simon.floery@gmx.at)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
/*
 * Initial main.c file generated by Glade. Edit as required.
 * Glade will not overwrite this file.
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <locale.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "calc_basic.h"
#include "galculator.h"
#include "display.h"
#include "config_file.h"
#include "general_functions.h"
#include "callbacks.h"
#include "interface.h"
#include "support.h"

#define MASK_NUMLOCK GDK_MOD2_MASK

GtkWidget 				*main_window;
char					dec_point=DEFAULT_DEC_POINT;
extern s_preferences	prefs;
s_current_status 		current_status = {0, 0, 0, 0};

void print_usage ()
{
	printf ("\n%s v%s, (c) 2002-2003 Simon Floery\n\n\
Usage: %s [options]\n\n\
options:\n\
(the GTK options)\n\
 -h, --help\t\toutput this usage message\n\
 -v, --version\t\tversion information\n\
galculator has no specific command line options any more but an extensive \
preferences dialog/configuration file system. Make any settings there.\n\n", \
PROG_NAME, PROG_VERSION, PROG_NAME);
}

int key_snooper (GtkWidget *grab_widget, GdkEventKey *event, gpointer func_data)
{
	/* the problem: NUMLOCK is as a modifier for gdk, like Shift or Control or Alt.
	 * If NUMLOCK is activated, an accelerator defined e.g. with CTRL-q won't work as they
	 * are caught as CTRL-NUMLOCK-q. As a matter of fact, we can't set the NUMLOCK modifier
	 * in GLADE. In order to not modify interface.c everytime the GUI changed, I catch here
	 * any key pressed and delete, if not 0-9 or decimal on keypad, the NUMLOCK flag.
	 * 0-9 and decimal on the keypad need the NUMLOCK flag, otherwise they are interpreted
	 * like Pgup and all the funny staff found on the deactivated keypad.
	 */

	//printf ("snooper: %i %s\n", event->keyval, gdk_keyval_name (event->keyval));
	if (((event->keyval < GDK_KP_0) || (event->keyval > GDK_KP_9)) && \
		(event->keyval != GDK_KP_Decimal)) event->state &= ~GDK_MOD2_MASK;
	return FALSE;
}
	
int main (int argc, char *argv[])
{
	char			*dec_point_string, *config_file_name;
	struct lconv 	*locale_settings;
	
	/*
	 * gtk_init runs (among other things) setlocale (LC_ALL, ""). Therefore we
	 * have to/can deal with i18n from this point.
	 * call gtk_init before getopts to get gtk options stripped (--display etc)
	 */
	
	gtk_init (&argc, &argv);
	
	if (argc > 1) {
		print_usage ();
		return EXIT_SUCCESS;
	}
	
	/* at first, get config file */
	config_file_name = g_strdup_printf ("%s/%s", getenv ("HOME"), CONFIG_FILE_NAME);
	config_file_read (config_file_name);
	g_free (config_file_name);
	
	/* as the following feature is out of scope of config_file.c, we do it here
	 * handle "remember display value on exit" and newline
	 */
	
	if (prefs.rem_display == FALSE) {	
		g_free (prefs.rem_value);
		prefs.rem_value = g_strdup (DEFAULT_REM_VALUE);
	}	
	
	/* get locale's decimal point */
	
	locale_settings = localeconv();
	if (strlen (locale_settings->decimal_point) != 1) \
		fprintf (stderr, "[%s] length of locale's decimal point is not supported: >%s<\n\
There might be problems when using galculator! %s\n", PROG_NAME, locale_settings->decimal_point, BUG_REPORT);
	else dec_point = locale_settings->decimal_point[0];
		
	//add_pixmap_directory (PACKAGE_DATA_DIR "/" PACKAGE "/pixmaps");
   	
	main_window = create_main_window ();
	
	gtk_window_set_title ((GtkWindow *)main_window, PROG_NAME);

	/* usually, only Shift, CTRL and ALT modifiers are paid attention to by accelerator code.
	 * add MOD2 (NUMLOCK allover the worl?) to the list. We have to do this for a working
	 * keypad.
	 */
	  
	gtk_accelerator_set_default_mod_mask (gtk_accelerator_get_default_mod_mask () | GDK_MOD2_MASK); 
	
	/* update "decimal point" button to locale's decimal point */
	
	dec_point_string = g_strdup_printf ("%c", dec_point);
	gtk_button_set_label ((GtkButton *) lookup_widget (main_window, "button_point"), \
		dec_point_string);
	g_free (dec_point_string);
							  
	/* prepare calc_basic */
	
	calc_tree_init (0);
	calc_rpn_init (0);

	statusbar_init (main_window);

	/* finally show what we put together. do this as late asap */
	gtk_widget_show (main_window);	
	
	/* main_windows has to be visible to get a nice and proper display */	
	display_init (main_window);
	
	/* apply changes */
	update_all (main_window, prefs);

	memory_update_tooltips ();

	// see function key_snooper for details
	gtk_key_snooper_install (key_snooper, NULL);

	gtk_main ();

	/* save changes to file */

	config_file_name = g_strdup_printf ("%s/%s", getenv ("HOME"), CONFIG_FILE_NAME);
	config_file_write (config_file_name, prefs);
	g_free (config_file_name);

	return EXIT_SUCCESS;
}
